/*
 *  Draco Econometrics
 *  Copyright (C) 2009 Approximatrix, LLC
 *  http://www.approximatrix.com
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.approximatrix.draco.application.plugin.ui.menu;

import java.awt.event.ActionEvent;
import java.io.IOException;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import com.approximatrix.draco.application.ApplicationManager;
import com.approximatrix.draco.application.MainWindowProvider;
import com.approximatrix.draco.application.plugin.IOPlugin;
import com.approximatrix.draco.application.plugin.Plugin;
import com.approximatrix.draco.application.plugin.ui.PluginErrorDialog;
import com.approximatrix.draco.application.plugin.ui.PluginOptionDialog;
import com.approximatrix.draco.application.plugin.ui.PluginWindow;
import com.approximatrix.draco.general.ui.GeneralWorker;
import com.approximatrix.draco.io.filters.PlainTextFilter;
import com.approximatrix.draco.io.filters.RubyFileFilter;
import com.approximatrix.draco.main.ui.MainWindow;

public class IOPluginMenuHandler extends AbstractPluginMenuHandler implements MainWindowProvider {

	private MainWindow parent = null;
	private boolean importing = false;
	
	private IOPluginMenuHandler(MainWindow parent) {
		super();
		this.parent = parent;
	}
	
	public IOPluginMenuHandler(MainWindow parent, boolean importing) {
		this(parent);
		this.importing = importing;
	}

	public void actionPerformed(ActionEvent e) {
		if(!(e.getSource() instanceof PluginMenuItem)) 
			return;
		
		PluginMenuItem pluginItem = (PluginMenuItem)e.getSource();
		IOPlugin topass = null;
		try {
			topass = (IOPlugin)pluginItem.getPlugin();
		} catch(ClassCastException cce) {
			new PluginErrorDialog(parent,cce).setVisible(true);
			return;
		}
		
		final IOPlugin worker = (IOPlugin)topass.createInstance();
		
		new GeneralWorker(this.getMainWindow(),
				"Initializing Plugin","Starting "+pluginItem.getText(),
				new GeneralWorker.WorkRoutine() {
					public void doWork() {
						
						try {
							
							worker.initialize(parent.getWindowTableModel());
							
							// Have to set the plugin action immediately following intialization
							if(IOPluginMenuHandler.this.importing)
								worker.setImport();
							else
								worker.setExport();
							
							String filename = IOPluginMenuHandler.this.requestFilename(worker);
							
							if(filename == null)
								return;
							
							worker.setFilename(filename);
							
							worker.executePreliminaryCode();
						} catch(Exception e) {
							
							worker.setFilename(null);
							
							final Exception tpe = e;
							SwingUtilities.invokeLater(new Runnable() {
								public void run() {
									new PluginErrorDialog(parent,tpe).setVisible(true);
								}
							});
							
						}
						
					}
				});
		
		if(worker.getFilename() == null) 
			return;
		
		// Display options
		new PluginOptionDialog(parent,worker.getOptions(),worker.getName());
		
		// Do the work
		new GeneralWorker(this.getMainWindow(),
				"Executing Plugin","Running "+pluginItem.getText(),
				new GeneralWorker.WorkRoutine() {
					public void doWork() {
						
						try {
							
							worker.execute();
							
						} catch(Exception e) {
							
							final Exception tpe = e;
							SwingUtilities.invokeLater(new Runnable() {
								public void run() {
									new PluginErrorDialog(parent,tpe).setVisible(true);
								}
							});
							
						}
						
					}
				});
		
		// Import the data if applicable
		if(worker.getModel() != null && this.importing) {
			if(!this.getMainWindow().isInitialWindow())
				ApplicationManager.RequestNewWindow(worker.getModel());
			else
				this.getMainWindow().setWindowTableModel(worker.getModel());
		} else if(this.importing)
			JOptionPane.showMessageDialog(this.getMainWindow(), "A valid data model was not created during import", "Import Error",JOptionPane.WARNING_MESSAGE);
			
	}

	public MainWindow getMainWindow() {
		return parent;
	}
	
	private String requestFilename(IOPlugin plugin) throws IOException {
		JFileChooser chooser = new JFileChooser();
		
		chooser.setDialogType(JFileChooser.SAVE_DIALOG);
		chooser.setFileFilter(plugin.getFileFilter());
		if(plugin.isImport()) {
			chooser.setDialogTitle("Select File to Import...");
			if (chooser.showOpenDialog(chooser) == JFileChooser.APPROVE_OPTION)
				return chooser.getSelectedFile().getCanonicalPath();
	    } else {
	    	chooser.setDialogTitle("Save As...");
	    	if (chooser.showSaveDialog(chooser) == JFileChooser.APPROVE_OPTION)
				return chooser.getSelectedFile().getCanonicalPath();
	    }
	    return null;
	}
	
}
