/*
 *  Draco Econometrics
 *  Copyright (C) 2006-2007 Approximatrix, LLC
 *  http://www.approximatrix.com
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.approximatrix.draco.fit;


import java.util.List;

import org.apache.commons.math.FunctionEvaluationException;
import org.apache.commons.math.optimization.GoalType;
import org.apache.commons.math.optimization.MultivariateRealOptimizer;
import org.apache.commons.math.optimization.OptimizationException;
import org.apache.commons.math.optimization.RealPointValuePair;
import org.apache.commons.math.optimization.direct.NelderMead;
import org.netlib.blas.DDOT;
import org.netlib.lapack.DGETRF;
import org.netlib.lapack.DGETRI;
import org.netlib.util.intW;

import com.approximatrix.draco.datamodel.StatTableModel;
import com.approximatrix.draco.fit.functions.*;
import com.approximatrix.draco.fit.statistics.ExtendedStatistic;
import com.approximatrix.draco.io.XMLio.AnalysisLoadStore;
import com.approximatrix.draco.math.MatrixSupport;
import com.approximatrix.draco.math.VectorSupport;
import com.approximatrix.draco.utility.ResultsUtilities;

public class SimpleMLEFit extends AbstractFit implements LSFit {
	
	private static boolean OVERRIDE_HESSIAN_SAFETY = true;
	
	private LikelihoodFunction evaluator = null;
	
	private String dependent_var = null;
	private String[] independent_var = null;
	
	private double[] parameters = null;
	private double[][] covariance_matrix = null;
	
	private Class evaluationFunction = null;
	private MultivariateRealOptimizer optimizer = null;
	
	private int[] use_row = null;
	
	/** The list of extended statistics to compute */
	private List<ExtendedStatistic> estats = null;
	
	/** Sets all data in the fit to the information loaded from
	 * a file
	 * 
	 * @param data a valid AnalysisLoadStore object that has been properly filled
	 */
	public void setLoadedAnalysis(AnalysisLoadStore data, StatTableModel model) {
		independent_var = data.getIndependentVariables();
		dependent_var = data.getDependentVariable();
		results = data.getResults();
		parameters = data.getCoef();
		covariance_matrix = data.getCovarianceMatrix();
		
		if(data.getAnalysisSubType() != null) {
			
			if(data.getAnalysisSubType().compareTo(LogitLikelihood.class.getName())==0)
				evaluator = new LogitLikelihood();
			else if(data.getAnalysisSubType().compareTo(ProbitLikelihood.class.getName())==0)
				evaluator = new ProbitLikelihood();
		}
		
		success = true;
	}
	
	public SimpleMLEFit(StatTableModel model) {
		super(model);
		// Nothing is done here - should only be called before a call to
		// setLoadedAnalysis.
	}
	
	public SimpleMLEFit(StatTableModel model,
			      String dependent,
			      String[] independent,
			      Class evaluationFunction) {
		this(model);
		this.dependent_var = dependent;
		this.independent_var = independent;
		this.evaluationFunction = evaluationFunction;
		
	}

	public LikelihoodFunction getFunction() {
		return evaluator;
	}

	/** Checks if an evaluator has been selected and the proper data is present in storage */
	private boolean checkAndInitializeEvaluator() {
		
		if(getFunction() == null && this.evaluationFunction != null) {
			if(this.evaluationFunction == LogitLikelihood.class)
				evaluator = new LogitLikelihood();
			else if(this.evaluationFunction == ProbitLikelihood.class)
				evaluator = new ProbitLikelihood();
			else {
				success = false;
				failure_message = "Unknown functional evaluation";
				return false;
			}	
		}
		
		if(!evaluator.isReady()) {
			use_row = determineUsable();
			double[][] independent_data = getModel().getDataDouble(use_row,getModel().getColumnNumber(independent_var));
			double[] dependent_data = getModel().getDataDouble(use_row,getModel().getColumnNumber(dependent_var));
			evaluator.setData(independent_data, dependent_data);
		}
		return true;
	}
	
	public String getResults() {
		if(results == null) {
			
			checkAndInitializeEvaluator();
			
			// Add some info about convergence
			addMessage("Final Likelihood: "+getFunction().evaluateLikelihood(getCoef()));
			addMessage("Optimizer scheme: "+optimizer.getClass().getSimpleName());
			addMessage("Number of iterations: "+Integer.toString(this.optimizer.getIterations()));
			
			StringBuffer text = null;
			String results_header = "Results of the "+evaluator.getLikelihoodType()+" MLE Regression Model\n";
			if(!success) {
				text = LSFitResultsGen.GenerateFailureResults(this,getFailureMessage(),results_header);
			} else {
				try {
					text = LSFitResultsGen.GenerateResults(this,results_header);
				} catch(Exception e) {
					
					this.failure_message = "Unintelligible results: "+e.getMessage()+"<br>Convergence probably did not occur.";
					if(!ResultsUtilities.USE_HTML)
						this.failure_message = this.failure_message.replaceAll("<br>", "\n");
					
					text = LSFitResultsGen.GenerateFailureResults(this,getFailureMessage(),results_header);
					success = false;
				}
				//print_special_options(text);
				//checkExtendedOutput(text);
			}
			results = text.toString();
			
			evaluator.free();
		}
		return results;
	}

	public double[][] getCovarianceMatrix() {
		if(covariance_matrix != null) return covariance_matrix;
		if(!success) return null;

		double[][] ls_coef = generate_Hessian_matrix();
		
		ls_coef = MatrixSupport.Negate(ls_coef);
		
		// Check if this can be invertedhessian matrix logit
		double rcond = MatrixSupport.GetRecipricolCondition(ls_coef);
		
		if(rcond > RunLS.MINIMUM_R_CONDITION || OVERRIDE_HESSIAN_SAFETY)
			covariance_matrix = MatrixSupport.Invert(ls_coef);
		else {
			covariance_matrix = MatrixSupport.Identity(parameters.length);
			this.addMessage("The covariance matrix could not be computed: the negated Hessian is not invertible");
		}
		
		return covariance_matrix;
	}

	public double[] getResiduals() {
		
		checkAndInitializeEvaluator();
		
		double[][] data = getModel().getDataDouble(use_row,getModel().getColumnNumber(independent_var));
		double[] results = getModel().getDataDouble(use_row,getModel().getColumnNumber(dependent_var));
		for(int i=0;i<results.length;i++) 
			results[i] = results[i] - evaluator.evaluatePDF(parameters, data[i]);
		
		getFunction().free();
		
		return results;
	}

	public double getResidualSumSqared() {
		double[] residuals = getResiduals();
		if(residuals == null) return 0;
		
		return DDOT.DDOT(residuals.length,residuals,1,residuals,1);
	}

	public double getTotalSumSqared() {
		double[] dep = getModel().getDataDouble(use_row,getModel().getColumnNumber(dependent_var));
		if(dep == null) return 0;
		
		return DDOT.DDOT(dep.length,dep,1,dep,1);
	}

	public double[] getCoef() {
		return parameters;
	}

	public String[] getCoefLabels() {
		
		if(parameters.length == independent_var.length) {
			return getIndependentVariables();
		}
		
		String[] names = new String[independent_var.length+1];
		names[0] = "Constant";
		for(int i=0;i<independent_var.length;i++) names[i+1] = independent_var[i];
		
		return names;
	}

	public double[] getCoefVariances() {
		double[] variances = MatrixSupport.GetDiagonal(getCovarianceMatrix());
		return variances;
	}

	public String[] getIndependentVariables() {
		return independent_var;
	}

	public String getDependentVariable() {
		return dependent_var;
	}

	public Double[] getFittedData() {
		Double[] result = new Double[getModel().getRowCount()];
		
		double[][] data = getModel().getDataDouble(use_row,getModel().getColumnNumber(independent_var));
		
		
		for(int i=0;i<getModel().getRowCount();i++)
			result[i] = null;
		
		checkAndInitializeEvaluator();
		
		for(int i=0;i<use_row.length;i++)
			result[use_row[i]] = evaluator.evaluateCDF(parameters, data[i]);

		evaluator.free();
		
		return result;
	}

	public Double[] getResidualData() {
		Double[] results = new Double[getModel().getRowCount()];
		for(int i=0;i<results.length;i++)
			results[i] = null;
		
		double[] residuals = getResiduals();
		for(int i=0;i<use_row.length;i++) {
			results[use_row[i]] = new Double(residuals[i]);
		}
		return results;
	}

	private int[] determineUsable() {
		String[] all_columns = new String[independent_var.length+1];
		all_columns[0] = dependent_var;
		for(int i=0;i<independent_var.length;i++)
			all_columns[i+1] = independent_var[i];
		return getModel().getNonNullRows(all_columns);
	}
	
	private boolean regress() {
		// Exit if the evaluator init fails
		if(!checkAndInitializeEvaluator()) return false;
		parameters = evaluator.initializeParameters();
		
		//optimizer = new MultiDirectional();
		optimizer = new NelderMead();
		((NelderMead)optimizer).setStartConfiguration(VectorSupport.Scale(0.05, parameters));
		try {
			RealPointValuePair result = optimizer.optimize(evaluator, GoalType.MAXIMIZE, parameters);
			success = true;
			parameters = result.getPointRef();
		} catch(FunctionEvaluationException fee) {
			failure_message = "An error occurred while attempting to evaluate the regression function.";
			success = false;
		} catch(OptimizationException oe) {
			failure_message = "The optimizer failed during calculations.";
			success = false;
		} catch(IllegalArgumentException iae) {
			failure_message = "An internal error has occurred (IllegalArgumentException) within the optimizer.";
			success = false;
		} finally {
			evaluator.free();
		}
		
		return success;
	}

	private String getFailureMessage() {
		if(failure_message != null)
			return failure_message;
		else
			return "Error in SimpleMLEFit algorithm";
	}

	/** Returns the variance of the residuals
	 * 
	 * @return double representing the residual variance
	 */
	private double getResidualVariance() {
		double denom = new Integer(use_row.length).doubleValue() -
		    new Integer(parameters.length).doubleValue();
		return getResidualSumSqared()/denom;
	}
	
	private double[][] generate_Hessian_matrix() {
		checkAndInitializeEvaluator();
		double[][] ret = evaluator.getHessian(parameters);
		evaluator.free();
		return ret;
	}

	public void perform() {
		if(this.success) {
			this.success = regress();
			
			// pre-calculate the covariance here - it could be expensive
			getCovarianceMatrix();
		}
	}
	
	/** This method computes and returns the extended statistics
	 * 
	 * @see com.approximatrix.draco.fit.LSFit#getExtendedStatistics()
	 */
	public List<ExtendedStatistic> getExtendedStatistics() {
		if(estats == null) return null;
		for(ExtendedStatistic stat : estats) {
			stat.setAnalysis(this);
			stat.compute();
		}
		return estats;
	}

	/** Sets the list of extended statistics
	 * @see com.approximatrix.draco.fit.LSFit#setExtendedStatistics(java.util.List)
	 */
	public void setExtendedStatistics(List<ExtendedStatistic> statistics) {
		estats = statistics;
	}
	
	/** Returns the current list of extended statistics without performing computations
	 * @see com.approximatrix.draco.fit.LSIterative#getExtendedStatisticsWithoutCompute()
	 */
	public List<ExtendedStatistic> getExtendedStatisticsWithoutCompute() {
		return estats;
	}
}
