/*
 *  Draco Econometrics
 *  Copyright (C) 2008-2009 Approximatrix, LLC
 *  http://www.approximatrix.com
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.approximatrix.draco.fit.functions;

import org.apache.commons.math.FunctionEvaluationException;
import org.apache.commons.math.analysis.MultivariateRealFunction;
import org.netlib.blas.DAXPY;

import com.approximatrix.draco.fit.RunLS;
import com.approximatrix.draco.math.VectorSupport;

public abstract class SimpleLikelihood implements LikelihoodFunction, MultivariateRealFunction {

	private static final double RELATIVE_PERTURBATION = 0.001;
	private static final double ABSOLUTE_PERTURBATION_BACKUP = 1.0E-8;
	
	protected double[][] independent;
	protected double[] dependent;
	
	protected SimpleLikelihood() {
		this.independent = null;
		this.dependent = null;
	}
	
	protected SimpleLikelihood(double[][] independent, double[] dependent) {
		this.setData(independent, dependent);
	}
	
	public double[] initializeParameters() {
		double[] results = RunLS.SimpleLS(independent, dependent);
		if(results == null) return null;
		return results;
		//return VectorSupport.VectorOfZeros(independent[0].length+1);
	}

	public double value(double[] parameters) throws FunctionEvaluationException,
			IllegalArgumentException {
		return evaluateLikelihood(parameters);
	}
	
	//public abstract double evaluateFunction(double[] parameters, double[] variables);
	
	public double evaluateLikelihood(double[] parameters) {
		double llh = 0.0;
		for(int i=0;i<dependent.length;i++)
			llh += evaluateLikelihoodRow(parameters,independent[i],dependent[i]);
	
		return llh;
	}
	
	public abstract double evaluateLikelihoodRow(double[] parameters, double[] independent, double dependent);
	
	protected double calculateLinearTerm(double[] parameters, double[] variables) {
		// For the constant to be included
		double[] complete_vars = new double[variables.length+1];
		complete_vars[0] = 1.0;
		for(int i=0;i<variables.length;i++)
			complete_vars[i+1] = variables[i];
		
		return VectorSupport.VectorElementSum(VectorSupport.MultiplyElements(parameters,complete_vars));
	}
	
	/** Disturbs the element of the vector, returning a new copy of the vector
	 * with a single perturbed element
	 * 
	 * @param parameters the vector to perturb
	 * @param index the index to perturb
	 * @return a new vector
	 */
	private double[] disturb(final double[] parameters, int index, boolean positive) {
		double[] ret = parameters.clone();
		if(positive)
			ret[index] += perturbation(parameters,index);
		else
			ret[index] -= perturbation(parameters,index);
		return ret;
	}
	
	/** Computes an acceptable perturbation for a specified parameter
	 * 
	 * @param parameters the vector of parameter values
	 * @param index the index to calculate the perturbation for
	 * @return a perturbation
	 */
	private double perturbation(final double[] parameters, int index) {
		double ret;
		if(parameters[index] == 0.0)
			ret = VectorSupport.Mean(parameters)*RELATIVE_PERTURBATION;
		else
			ret = parameters[index]*RELATIVE_PERTURBATION;
		
		if(Math.abs(ret) < 1.0E-15)
			ret = ABSOLUTE_PERTURBATION_BACKUP;
		
		return ret;
	}
	
	public double[][] getHessian(double[] parameters) {
		double[][] return_val = new double[parameters.length][parameters.length];
		
		for(int i = 0;i<parameters.length;i++) {

			for(int j = i;j<parameters.length;j++) {
				
				double[] disturbed_pipj = disturb(disturb(parameters,i,true),j,true);
				double[] disturbed_pimj = disturb(disturb(parameters,i,true),j,false);
				double[] disturbed_mipj = disturb(disturb(parameters,i,false),j,true);
				double[] disturbed_mimj = disturb(disturb(parameters,i,false),j,false);
								
				return_val[i][j] = this.evaluateLikelihood(disturbed_pipj);
				return_val[i][j] -= this.evaluateLikelihood(disturbed_pimj);
				return_val[i][j] -= this.evaluateLikelihood(disturbed_mipj);
				return_val[i][j] += this.evaluateLikelihood(disturbed_mimj);
				
				return_val[i][j] = return_val[i][j] / (4.0*this.perturbation(parameters, i)*this.perturbation(parameters, j));
			}
		}
		
		// Only the top was filled in, copy the top to the bottom
		for(int i=0;i<parameters.length;i++) {
			for(int j=i+1;j<parameters.length;j++) {
				return_val[j][i] = return_val[i][j];
			}
		}
		return return_val;
	}

	public void free() {
		this.independent = null;
		this.dependent = null;
	}

	public boolean isReady() {
		return this.independent != null && this.dependent != null;
	}

	public void setData(double[][] independent, double[] dependent) {
		this.independent = independent;
		this.dependent = dependent;
	}
}
