package com.approximatrix.draco.math.functions;

import org.apache.commons.math.MathException;
import org.apache.commons.math.special.Erf;

/** Implementation of the Error Function (erf) and related
 * functions.
 * 
 * @author jarmstrong
 *
 */
public class ErrorFunction {

	/** Returns the error function of value.  Based on
	 * call to SSJ.
	 * 
	 * @param value the input to the error function
	 * @return erf(value)
	 */
	public static double Erf(double value) {
		try {
			return Erf.erf(value);
		} catch(MathException me) {
			return 0;
		}
	}
	
	/** Returns the complimentary error function of value.  
	 * Based on call to SSJ.
	 * 
	 * @param value the input to the error function
	 * @return erf(value)
	 */
	public static double Erfc(double value) {
		return 1.0-Erf(value);
	}

	private static final double TOL = 1.0E-8;
	private static final int MAX_ITERATIONS = 100;
	
	/** Computes the inverse error function by using the error function itself
	 * and iteratively locating the answer.  
	 * 
	 * This algorithm is adapted from the corresponding Octave 
	 * (http://www.gnu.org/software/octave/) function.  The original form is 
	 * by Kurt Hornik <Kurt.Hornik@wu-wien.ac.at> and licensed under the GNU
	 * General Public Licence Version 3.
	 *   
	 * @param x
	 * @return
	 * @throws IllegalArgumentException
	 */
	public static double InvErf(double x) throws IllegalArgumentException {
		/* Copyright (C) 1995, 1996, 1997, 1999, 2000, 2002, 2004, 2005, 2006,
         *     2007 Kurt Hornik
         *
         * This file is part of Octave.
         *
         * Octave is free software; you can redistribute it and/or modify it
         * under the terms of the GNU General Public License as published by
         * the Free Software Foundation; either version 3 of the License, or (at
         * your option) any later version.
         *
         * Octave is distributed in the hope that it will be useful, but
         * WITHOUT ANY WARRANTY; without even the implied warranty of
         * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
         * General Public License for more details.
         *
         * You should have received a copy of the GNU General Public License
         * along with Octave; see the file COPYING.  If not, see
         * <http://www.gnu.org/licenses/>.
         */
		
		if(x < -1.0 || x > 1.0)
			throw new IllegalArgumentException("InvErf requires an input between -1 and 1");
		
		if(x == 1)
			return Double.POSITIVE_INFINITY;
		else if(x == -1)
			return Double.NEGATIVE_INFINITY;
		
		double s = Math.sqrt(Math.PI) / 2.0;
		double z_old = 1.0;
		double z_new = Math.sqrt(-1.0*Math.log(1.0 - Math.abs (x))) * Math.signum(x);
		
		int iterations = 0;
		try {
			while (Math.abs(Erf.erf(z_new) - x) > TOL * Math.abs (x) && iterations < MAX_ITERATIONS) {
				z_old = z_new;
				z_new = z_old - (Erf.erf(z_old) - x) * Math.exp(z_old*z_old) * s;
			}
		} catch(MathException me) {
			//pass
		}
		return z_new;
	}
	
	/** Provides a clipping version of the inverse error function.  In other words,
	 * the value is clipped to make sure the inputs to the actual inverse erf calc
	 * are safe.
	 * 
	 * @param x the probability, which is clipped outside of [-1,1]
	 * @return inverse erf result
	 */
	public static double InvErfClip(double x) {
		return InvErf(Math.max(-1.0, Math.min(1.0,x)));
	}

}
